/*

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish and distribute.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
  AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Copyright 1999 -   Laurent Balmelli
                     Laboratory For Audio-Visual Communications 
                     Ecole Polytechnique Federale de Lausanne, Switzerland

                     also:
                     Mathematics for Communication, Bell Laboratories
                     Lucent Technologies, USA

  contact, bugs:     Laurent.Balmelli@epfl.ch
                     balmelli@research.bell-labs.com
                     
  PLEASE DO NOT REMOVE THIS COPYRIGHT NOTICE. 
                     
 */

import java.lang.Math;
import java.awt.*;
import java.awt.image.*;
import java.awt.Toolkit;
//import java.lang.*;
import java.lang.reflect.Array;

// TODO
// Output of an array of signals, instead of one FilteredSignal


 class Filter 
{

	
	public double[] FilteredSignals;
	
	public final static int H0 = 0;
	public final static int H1 = 1;
	public final static int G0 = 2;
	public final static int G1 = 3;	

	private double h0[];
	private double h1[];
	private double g0[];
	private double g1[];

	private int H0set;
	private int H1set; 
	private int G0set;
	private int G1set;
	private int FilteredSignalsSet;

	private int FilterDelay;

	private double AlphaBuffer[];
	private double RedBuffer[];
	private double GreenBuffer[];
	private double BlueBuffer[];
	private ImageConverter imageData;
	private int ImageWidth,
		    ImageHeight;

	private double AlphaBufferProjection[];
	private double RedBufferProjection[];
	private double GreenBufferProjection[];
	private double BlueBufferProjection[];	

	public Filter()
	{ 
	   H0set = 0;  H1set = 0;  G0set = 0;  G1set = 0; 
	   
	   //FilteredSignalsSet = 1;	   
 	   //FilteredSignals = new double[FilteredSignalsSet];

	   //FilteredSignals[0] = new double[1];
	   //FilteredSignal[0][1] = 0.0;	
		
		
	}

	public Filter(Image img, int width, int height)
	{
	   H0set = 0;  H1set = 0;  G0set = 0;  G1set = 0; 
	   
	   imageData = new ImageConverter(img,width,height);
	   ImageWidth = width;
	   ImageHeight = height;
	
	   int[] pixels = imageData.getIntBuffer();

	   //
	   // Convert the image into 4 buffers of double
	   //

	   //System.out.println("Converting image to double buffers...");
	
	   AlphaBuffer = imageData.getDoubleBufferComponent(pixels,
							    imageData.ALPHA);
	   RedBuffer   = imageData.getDoubleBufferComponent(pixels,
							    imageData.RED);
	   GreenBuffer = imageData.getDoubleBufferComponent(pixels,
						            imageData.GREEN);
	   BlueBuffer  = imageData.getDoubleBufferComponent(pixels,
							    imageData.BLUE);	
	   //System.out.println("done.");
	}

	public void setFilter(int type, double[] filter)
	{
	   if ( type == H0 ) {
		h0 = new double[filter.length];
		for ( int i = 0; i < filter.length; i++ )
			h0[i] = filter[i];
		H0set = 1;
	   }

	   if ( type == H1 ) {
		h1 = new double[filter.length];
		for ( int i = 0; i < filter.length; i++ )
			h1[i] = filter[i];
		H1set = 1;
	   }

           if ( type == G0 ) {
		g0 = new double[filter.length];
		for ( int i = 0; i < filter.length; i++ )
			g0[i] = filter[i];
		G0set = 1;
	   }

           if ( type == G1 ) {
		g1 = new double[filter.length];
		for ( int i = 0; i < filter.length; i++ )
			g1[i] = filter[i];
		G1set = 1;
	   }
	   
	   //printData("Filter value is (length = " + filter.length + ") ",filter);
	}

	public void setFilterDelay(int delay)
	{
		FilterDelay = delay;
	}		
	
	public void testTwoChannel(double signal[])
	{
		int slength = signal.length;
		double recons[] = new double[slength];
		double low[]; double high[];
		int i,j;

		//printData("Original signal ",signal);

		//printData("H0",h0);
		low = conv(h0,signal);
		low = downSample(low,2);
		//printData("Transform coefficients low",low);	

		//printData("H1",h1);
		high = conv(h1,signal);
		high = downSample(high,2);
		//printData("Transform coefficients high",high);

		//printData("g0",g0);
		low = upSample(low,2);
		low = conv(g0,low);
		//printData("Projection V0",low);

		//printData("g1",g1);
		high = upSample(high,2);
		high = conv(g1,high);
		//printData("Projection W0",high);

		// computing reconstructed signal
	
		for (i=0; i< low.length; i++) recons[i] = low[i]+high[i];

		//printData("Reconstructed signal",recons);

		// Store Coarse Resolution in FilteredSignal
		
		FilteredSignals = new double[recons.length];

		for ( i=0; i < recons.length; i++ ) FilteredSignals[i] = recons[i];	


	}

	//
	// 	ANALYSIS FOR 2D SIGNALS 
	//
	public int[] analysis(int width, int height,int scale, int level)
	{
	
		//
		// Filter Bank
	
		int i;
		
		//System.out.println("Analysis filter bank has started.");

		double low[],high[];
		int half_width = width/2;
		int half_height = height/2;

		//System.out.println("Filtering red channel...");
		
		for ( i=0; i < height; i++) {
		   low = conv(h0,RedBuffer,i*width,width);		   
		   high = conv(h1,RedBuffer,i*width,width);
		   low = downSample(low,2);
		   high = downSample(high,2);
   
	           System.arraycopy(low,0,RedBuffer,
		   		    i*width,half_width);
		   System.arraycopy(high,0,RedBuffer,
		   		    (i*width+half_width),half_width);
		}  

		RedBuffer = imageData.transposeBuffer(RedBuffer,width,height);
		
		for ( i=0; i < width; i++) {
		   low = conv(h0,RedBuffer,i*height,height);
		   high = conv(h1,RedBuffer,i*height,height);
		   low = downSample(low,2);
		   high = downSample(high,2);
	
	           System.arraycopy(low,0,RedBuffer,
				    i*height,half_height);
		   System.arraycopy(high,0,RedBuffer,
				    (i*height+half_height),half_height);		
		}
		
		RedBuffer = imageData.transposeBuffer(RedBuffer,height,width);

		//System.out.println("done.");
		
		//System.out.println("Filtering green channel...");
		for ( i=0; i < height; i++) {
		   low = conv(h0,GreenBuffer,i*width,width);
		   high = conv(h1,GreenBuffer,i*width,width);
		   low = downSample(low,2);
		   high = downSample(high,2);
   
	           System.arraycopy(low,0,GreenBuffer,
				    i*width,half_width);
		   System.arraycopy(high,0,GreenBuffer,
				    (i*width+half_width),half_width);
		}  

		GreenBuffer = imageData.transposeBuffer(GreenBuffer,width,height);

		for ( i=0; i < width; i++) {
		   low = conv(h0,GreenBuffer,i*height,height);
		   high = conv(h1,GreenBuffer,i*height,height);
		   low = downSample(low,2);
		   high = downSample(high,2);
	
	           System.arraycopy(low,0,GreenBuffer,
				    i*height,half_height);
		   System.arraycopy(high,0,GreenBuffer,
				    (i*height+half_height),half_height);		
		}

		GreenBuffer = imageData.transposeBuffer(GreenBuffer,height,width);
		
		//System.out.println("done.");
		
		//System.out.println("Filtering blue channel...");
		for ( i=0; i < height; i++) {
		   low = conv(h0,BlueBuffer,i*width,width);
		   high = conv(h1,BlueBuffer,i*width,width);
		   low = downSample(low,2);
		   high = downSample(high,2);
   
	           System.arraycopy(low,0,BlueBuffer,
				    i*width,half_width);
		   System.arraycopy(high,0,BlueBuffer,
				    (i*width+half_width),half_width);
		}  

		BlueBuffer = imageData.transposeBuffer(BlueBuffer,width,height);

		for ( i=0; i < width; i++) {
		   low = conv(h0,BlueBuffer,i*height,height);
		   high = conv(h1,BlueBuffer,i*height,height);
		   low = downSample(low,2);
		   high = downSample(high,2);
	
	           System.arraycopy(low,0,BlueBuffer,
				    i*height,half_height);
		   System.arraycopy(high,0,BlueBuffer,
				    (i*height+half_height),half_height);		
		}
		BlueBuffer = imageData.transposeBuffer(BlueBuffer,height,width);
		
		//System.out.println("done.");

		//
		// convert the four buffer of double into an single buffer of int
		//
		//System.out.println("Converting buffer to image...");

		int inverse[] = imageData.DoubleBufferToIntBuffer(AlphaBuffer,
								  RedBuffer,
								  GreenBuffer,
								  BlueBuffer,
								  scale,
								  level);
		//System.out.println("done.");

		return inverse;
		
	}

	public int[] analysis(int offsetx, int offsety, int twidth, int theight,int scale, int level)
	{
	
		//
		// Filter Bank
	
		int i;
		int width = twidth-offsetx;	int height = theight-offsety;

		int OffsetImageX = ImageWidth-width,
		    OffsetImageY = ImageHeight-height;

		int OffsetTransposeX = offsety*ImageWidth;
		int OffsetTransposeY = offsetx*ImageHeight;

		//System.out.println("Analysis filter bank has started.");

		double low[],high[];
		int half_width = width/2;
		int half_height = height/2;

		//System.out.println("Filtering red channel...");
		
		for ( i=0; i < height; i++) {
		   low = conv(h0,RedBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),width);		   
		   	   high = conv(h1,RedBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),width);
		   low = downSample(low,2);
		   high = downSample(high,2);
   
	           System.arraycopy(low,0,RedBuffer,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width);
		   System.arraycopy(high,0,RedBuffer,
		   		    (OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width),half_width);
		}  

		RedBuffer = imageData.transposeBuffer(RedBuffer,ImageWidth,ImageHeight);
		
		for ( i=0; i < width; i++) {
		   low = conv(h0,RedBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),height);
		   high = conv(h1,RedBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),height);
		   low = downSample(low,2);
		   high = downSample(high,2);
	
	           System.arraycopy(low,0,RedBuffer,
				    OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height);
		   System.arraycopy(high,0,RedBuffer,
				    (OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height),half_height);		
		}
		
		RedBuffer = imageData.transposeBuffer(RedBuffer,ImageHeight,ImageWidth);

		//System.out.println("done.");
		
		//System.out.println("Filtering green channel...");

		for ( i=0; i < height; i++) {
		   low = conv(h0,GreenBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),width);		   
		           high = conv(h1,GreenBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),width);
		   low = downSample(low,2);
		   high = downSample(high,2);
   
	           System.arraycopy(low,0,GreenBuffer,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width);
		   System.arraycopy(high,0,GreenBuffer,
		   		    (OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width),half_width);
		}  

		GreenBuffer = imageData.transposeBuffer(GreenBuffer,ImageWidth,ImageHeight);
		
		for ( i=0; i < width; i++) {
		   low = conv(h0,GreenBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),height);
		   high = conv(h1,GreenBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),height);
		   low = downSample(low,2);
		   high = downSample(high,2);
	
	           System.arraycopy(low,0,GreenBuffer,
				    OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height);
		   System.arraycopy(high,0,GreenBuffer,
				    (OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height),half_height);		
		}
		
		GreenBuffer = imageData.transposeBuffer(GreenBuffer,ImageHeight,ImageWidth);

		//System.out.println("done.");

		//System.out.println("Filtering blue channel...");

 		for ( i=0; i < height; i++) {
		   low = conv(h0,BlueBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),width);		   
		   high = conv(h1,BlueBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),width);
		   low = downSample(low,2);
		   high = downSample(high,2);
   
	           System.arraycopy(low,0,BlueBuffer,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width);
		   System.arraycopy(high,0,BlueBuffer,
		   		    (OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width),half_width);
		}  

		BlueBuffer = imageData.transposeBuffer(BlueBuffer,ImageWidth,ImageHeight);
		
		for ( i=0; i < width; i++) {
		   low = conv(h0,BlueBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),height);
		   high = conv(h1,BlueBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),height);
		   low = downSample(low,2);
		   high = downSample(high,2);
	
	           System.arraycopy(low,0,BlueBuffer,
				    OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height);
		   System.arraycopy(high,0,BlueBuffer,
				    (OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height),half_height);		
		}
		
		BlueBuffer = imageData.transposeBuffer(BlueBuffer,ImageHeight,ImageWidth);
		
		//System.out.println("done.");

		//
		// convert the four buffer of double into an single buffer of int
		//
		//System.out.println("Converting buffer to image...");

		int inverse[] = imageData.DoubleBufferToIntBuffer(AlphaBuffer,
								  RedBuffer,
								  GreenBuffer,
								  BlueBuffer,
								  scale,
								  level);
		//System.out.println("done.");

		return inverse;
		
	}

 
	// SYTNTHESIS

	public final static int PR  = 0;		// Ensures perfect reconstruction
	public final static int L   = 1;
	public final static int H   = 2;			

	public int[] synthesis(int width, int height, int Rtype1, int Rtype2,int scale, int level)
	{
		
		//
		// Filter Bank

		int i,j;
			
		//System.out.println("Synthesis filter bank has started.");

		double low[],high[];
		double reconstructed[];
		int half_width = width/2;
		int half_height = height/2;

		//System.out.println("Filtering red channel...");
		
		RedBuffer = imageData.transposeBuffer(RedBuffer,width,height);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(RedBuffer,i*height,half_height,2);
		   high = upSample(RedBuffer,i*height+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++) {
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      	//System.out.print("Warning: reconstructed value is out of bound.");
			   }		   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }
		  
	           System.arraycopy(reconstructed,0,RedBuffer,
		   		    i*height,height);
		}  
		reconstructed = new double[width];		

		RedBuffer = imageData.transposeBuffer(RedBuffer,height,width);
		
		for ( i=0; i < height; i++) {
		   low = upSample(RedBuffer,i*width,half_width,2);
		   high = upSample(RedBuffer,i*width+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++) {
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
                           }
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,RedBuffer,
		   		    i*width,width);
		
		}

		//System.out.println("done.");

		//System.out.println("Filtering green channel...");
		
		GreenBuffer = imageData.transposeBuffer(GreenBuffer,width,height);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(GreenBuffer,i*height,half_height,2);
		   high = upSample(GreenBuffer,i*height+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
 		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++) {
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      	//System.out.print("Warning: reconstructed value is out of bound.");
			   }
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }		   
		
	           System.arraycopy(reconstructed,0,GreenBuffer,
		   		    i*height,height);
		}  
		reconstructed = new double[width];		

		GreenBuffer = imageData.transposeBuffer(GreenBuffer,height,width);
		
		for ( i=0; i < height; i++) {
		   low = upSample(GreenBuffer,i*width,half_width,2);
		   high = upSample(GreenBuffer,i*width+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++) {
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];

		              //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
			   }
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,GreenBuffer,
		   		    i*width,width);
		
		}

		//System.out.println("done.");

		//System.out.println("Filtering blue channel...");
		
		BlueBuffer = imageData.transposeBuffer(BlueBuffer,width,height);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(BlueBuffer,i*height,half_height,2);
		   high = upSample(BlueBuffer,i*height+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++) {
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
			   }
		   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,BlueBuffer,
		   		    i*height,height);
		}  
		reconstructed = new double[width];		

		BlueBuffer = imageData.transposeBuffer(BlueBuffer,height,width);
		
		for ( i=0; i < height; i++) {
		   low = upSample(BlueBuffer,i*width,half_width,2);
		   high = upSample(BlueBuffer,i*width+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++) {
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
			   }

			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,BlueBuffer,
		   		    i*width,width);
		
		}

		//System.out.println("done.");
		
		//
		// convert the four buffer of double into an single buffer of int
		//
		//System.out.println("Converting buffer to image...");

		int inverse[] = imageData.DoubleBufferToIntBuffer(AlphaBuffer,
								  RedBuffer,
								  GreenBuffer,
								  BlueBuffer,
								  scale,
								  level);
		//System.out.println("done.");

		return inverse;
		
	}		

	
	public int[] synthesis(int offsetx, int offsety, int twidth, int theight, 
			       int Rtype1, int Rtype2,int scale, int level)
	{
		
		//
		// Filter Bank
		
		int width = twidth-offsetx;	int height = theight-offsety;

		int OffsetImageX = ImageWidth-width,
		    OffsetImageY = ImageHeight-height;

		int OffsetTransposeX = offsety*ImageWidth;
		int OffsetTransposeY = offsetx*ImageHeight;

		int i,j;
			
		//System.out.println("Synthesis filter bank has started.");

		double low[],high[];
		double reconstructed[];
		int half_width = width/2;
		int half_height = height/2;

		//System.out.println("Filtering red channel...");
		
		RedBuffer = imageData.transposeBuffer(RedBuffer,ImageWidth,ImageHeight);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(RedBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height,2);
		   high = upSample(RedBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++) {
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];
 			 
			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
			   }		   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   } 	   	   
		
	           System.arraycopy(reconstructed,0,RedBuffer,
		   		    OffsetTransposeY+offsety+i*(height+OffsetImageY),theight);
		}  
		reconstructed = new double[width];		

		RedBuffer = imageData.transposeBuffer(RedBuffer,ImageHeight,ImageWidth);
		
		for ( i=0; i < height; i++) {
		   low = upSample(RedBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width,2);
		   high = upSample(RedBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++) {
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];
			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");				
				   }		   	   
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,RedBuffer,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),twidth);
		
		}

		//System.out.println("done.");

		//System.out.println("Filtering green channel...");
		
		GreenBuffer = imageData.transposeBuffer(GreenBuffer,ImageWidth,ImageHeight);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(GreenBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height,2);
		   high = upSample(GreenBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
 
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++) {
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];	

			       //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      	//System.out.print("Warning: reconstructed value is out of bound.");
			   }
	   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,GreenBuffer,
		   		    OffsetTransposeY+offsety+i*(height+OffsetImageY),theight);
		}  
		reconstructed = new double[width];		

		GreenBuffer = imageData.transposeBuffer(GreenBuffer,ImageHeight,ImageWidth);
		
		for ( i=0; i < height; i++) {
		   low = upSample(GreenBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width,2);
		   high = upSample(GreenBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++) {
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];	

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
			  }	   	   
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,GreenBuffer,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),twidth);
		
		}

		//System.out.println("done.");

		//System.out.println("Filtering blue channel...");
		
		BlueBuffer = imageData.transposeBuffer(BlueBuffer,ImageWidth,ImageHeight);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(BlueBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height,2);
		   high = upSample(BlueBuffer,OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++) {
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];	

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
	   	   	   }
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,BlueBuffer,
		   		    OffsetTransposeY+offsety+i*(height+OffsetImageY),theight);
		}  
		reconstructed = new double[width];		

		BlueBuffer = imageData.transposeBuffer(BlueBuffer,ImageHeight,ImageWidth);
		
		for ( i=0; i < height; i++) {
		   low = upSample(BlueBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width,2);
		   high = upSample(BlueBuffer,OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++) {
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];

			      //if ( ((low[j]+high[j]) > 255) || ((low[j]+high[j]) < 0))
			      //	System.out.print("Warning: reconstructed value is out of bound.");
			  }
		   	   
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,BlueBuffer,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),twidth);
		
		}

		//System.out.println("done.");
		
		//
		// convert the four buffer of double into an single buffer of int
		//
		//System.out.println("Converting buffer to image...");

		int inverse[] = imageData.DoubleBufferToIntBuffer(AlphaBuffer,
								  RedBuffer,
								  GreenBuffer,
								  BlueBuffer,
								  scale,
								  level);
		//System.out.println("done.");

		return inverse;
		
	}


	public int[] projection(int offsetx, int offsety, int regionWidth, int regionHeight,
			        int Rtype1, int Rtype2, int level,int scale)
	{

	   // Copy the coefficients into the projection buffer
		
	   RedBufferProjection = new double[ImageWidth*ImageHeight];
	   System.arraycopy(RedBuffer,0,RedBufferProjection,0,
		            ImageWidth*ImageHeight);

	   GreenBufferProjection = new double[ImageWidth*ImageHeight];
	   System.arraycopy(GreenBuffer,0,GreenBufferProjection,0,
		            ImageWidth*ImageHeight);

	   BlueBufferProjection = new double[ImageWidth*ImageHeight];
	   System.arraycopy(BlueBuffer,0,BlueBufferProjection,0,
			    ImageWidth*ImageHeight);

	   int twidth = regionWidth;  int theight = regionHeight;

	   //System.out.println("Recovering projection from "+level+" levels.");

	   for ( int k = 0; k < level; k++ )
	   {	
		//
		// Filter Bank
		
		int width = twidth-offsetx;	int height = theight-offsety;

		int OffsetImageX = ImageWidth-width,
		    OffsetImageY = ImageHeight-height;

		int OffsetTransposeX = offsety*ImageWidth;
		int OffsetTransposeY = offsetx*ImageHeight;

		int i,j;
			
		//System.out.println("Synthesis filter bank has started.");

		double low[],high[];
		double reconstructed[];
		int half_width = width/2;
		int half_height = height/2;

		//System.out.println("Filtering red channel...");
		
		RedBufferProjection = imageData.transposeBuffer(RedBufferProjection,ImageWidth,ImageHeight);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(RedBufferProjection,OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height,2);
		   high = upSample(RedBufferProjection,OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];		   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   } 	   	   
		
	           System.arraycopy(reconstructed,0,RedBufferProjection,
		   		    OffsetTransposeY+offsety+i*(height+OffsetImageY),theight);
		}  
		reconstructed = new double[width];		

		RedBufferProjection = imageData.transposeBuffer(RedBufferProjection,ImageHeight,ImageWidth);
		
		for ( i=0; i < height; i++) {
		   low = upSample(RedBufferProjection,OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width,2);
		   high = upSample(RedBufferProjection,OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];		   	   
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,RedBufferProjection,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),twidth);
		
		}

		//System.out.println("done.");

		//System.out.println("Filtering green channel...");
		
		GreenBufferProjection = imageData.transposeBuffer(GreenBufferProjection,ImageWidth,ImageHeight);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(GreenBufferProjection,OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height,2);
		   high = upSample(GreenBufferProjection,OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
 
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];		   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }		   	   
		
	           System.arraycopy(reconstructed,0,GreenBufferProjection,
		   		    OffsetTransposeY+offsety+i*(height+OffsetImageY),theight);
		}  
		reconstructed = new double[width];		

		GreenBufferProjection = imageData.transposeBuffer(GreenBufferProjection,ImageHeight,ImageWidth);
		
		for ( i=0; i < height; i++) {
		   low = upSample(GreenBufferProjection,OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width,2);
		   high = upSample(GreenBufferProjection,OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];		   	   
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,GreenBufferProjection,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),twidth);
		
		}

		//System.out.println("done.");

		//System.out.println("Filtering blue channel...");
		
		BlueBufferProjection = imageData.transposeBuffer(BlueBufferProjection,ImageWidth,ImageHeight);
		reconstructed = new double[height];

		for ( i=0; i < width; i++) {
		   low = upSample(BlueBufferProjection,OffsetTransposeY+offsety+i*(height+OffsetImageY),half_height,2);
		   high = upSample(BlueBufferProjection,OffsetTransposeY+offsety+i*(height+OffsetImageY)+half_height,half_height,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);
	
		   // reconstructing orginal signal
		   
		   switch ( Rtype1 )
		   {
			case PR:
		   	   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j]+high[j];		   	   
			break;

			case L:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = low[j];
			break;

			case H:
			   for ( j=0; j < height; j++)
		              reconstructed[(j+FilterDelay)%height] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,BlueBufferProjection,
		   		    OffsetTransposeY+offsety+i*(height+OffsetImageY),theight);
		}  
		reconstructed = new double[width];		

		BlueBufferProjection = imageData.transposeBuffer(BlueBufferProjection,ImageHeight,ImageWidth);
		
		for ( i=0; i < height; i++) {
		   low = upSample(BlueBufferProjection,OffsetTransposeX+offsetx+i*(width+OffsetImageX),half_width,2);
		   high = upSample(BlueBufferProjection,OffsetTransposeX+offsetx+i*(width+OffsetImageX)+half_width,half_width,2);
		   low = conv(g0,low);		   
		   high = conv(g1,high);

		   // reconstructing orginal signal
		   
		   switch ( Rtype2 )
		   {
			case PR:
		   	   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j]+high[j];		   	   
			break;

			case L:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = low[j];
			break;

			case H:
			   for ( j=0; j < width; j++)
		              reconstructed[(j+FilterDelay)%width] = high[j];
			break;
		   }	   	   
		
	           System.arraycopy(reconstructed,0,BlueBufferProjection,
		   		    OffsetTransposeX+offsetx+i*(width+OffsetImageX),twidth);
		
		}

		//System.out.println("done.");

		twidth *= 2;	theight *=2;
		
	     }

	     //
	     // convert the four buffer of double into an single buffer of int
	     //
	
	   //System.out.println("Converting buffer to image...");

	     int inverse[] = imageData.DoubleBufferToIntBuffer(AlphaBuffer,
		   					       RedBufferProjection,
							       GreenBufferProjection,
							       BlueBufferProjection,
							       scale,level);
	     //System.out.println("done.");

	     return inverse;
		
	}

	public void clearBand(int orgx, int orgy, int width, int height) {
	  
	  int shifty = orgy*ImageWidth,
	      boundx = width-orgx,
	      boundy = height-orgy; 

	  for ( int i= 0; i < boundy; i++ ) {
	    for ( int j = 0; j < boundx; j++ ) {

	      RedBuffer[shifty+orgx+i*ImageWidth+j]   = 0.0; 
	      GreenBuffer[shifty+orgx+i*ImageWidth+j] = 0.0;
	      BlueBuffer[shifty+orgx+i*ImageWidth+j]  = 0.0;	

	    }
	  }
         }
	

	// --------------------------------------------------------------
	// Private Methods
	// all the function to implement a filter bank

	// --------------------
	// Circular convolution 
	// --------------------
	private double[] conv(double filter[], double signal[])
	{
		int ssize = signal.length,
		    fsize = filter.length;
		
		double result[] = new double[ssize];
		
		for ( int i = 0; i < ssize; i++)
  		{	
    		   result[i] = 0.0;

    		   for ( int j = fsize-1; j >=0; j--)
      		   {
			result[i] += filter[j]*signal[(i+j)%ssize];
      		   }
  		}

	   	return result;
	}

	// --------------------
	// Circular convolution
	// 	can compute convolution from offset to a certain extend given
	//	by length.
	// --------------------
	public double[] conv(double filter[], double signal[],
			     int offset, int length)
	{
		int fsize = filter.length;
		
		double result[] = new double[length];
		
		for ( int i = 0; i < length; i++)
  		{	
    		   result[i] = 0.0;

    		   for ( int j = fsize-1; j >=0; j--)
      		   {
			//result[i] += filter[j]*signal[(offseti+j)%length];
			result[i] += filter[j]*signal[offset+(i+j)%length];
      		   }
  		}

	   	return result;

	}
	
	// ------------------
	// Downsample	
	// ------------------

	private double[] downSample( double signal[], int fac )
	{
		int length = (signal.length-(signal.length%fac))/fac;	

		double result[] = new double[length];

		for ( int i = 0; i < length; i++)
  		   result[i] = signal[fac*i];
		
		return result;
	}

	// ------------------
	// Upsample
	// ------------------

	private double[] upSample( double signal[], int fac )
	{
		int length = signal.length*fac;

  		double result[] = new double[length];

		for ( int i = 0; i < signal.length; i++) {
    		   result[fac*i] = signal[i];

    		for (int j = 1; j < fac-1; j++) 
      		   result[fac*i + j ] = 0.0;

  		}

  		return result;
	}

	private double[] upSample( double signal[], int offset, int lsignal, int fac )
	{
		// copy in temporary buffer the part to upsample

		double buffer[] = new double[lsignal];
		System.arraycopy(signal,offset,buffer,0,lsignal);

		int length = lsignal*fac;

  		double result[] = new double[length];

		for ( int i = 0; i < lsignal; i++) {
    		   result[fac*i] = buffer[i];

    		for (int j = 1; j < fac-1; j++) 
      		   result[fac*i + j ] = 0.0;
  		}
		
		//System.out.println("Length = "+result.length);
  		return result;
	}	

	// --------------------------------------------------------------
	// Debug method

	private void printData( String text, double[] data )
	{
	    //System.out.println(text);
	    //System.out.print("\t"); 
	    //System.out.flush();
		for (int i = 0; i < data.length; i++ ) {
		   System.out.print(data[i]);
		   System.out.print(" ");
		   System.out.flush();
		}
		System.out.println();		
	}

}

