/*

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish and distribute.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
  AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Copyright 1999 -   Laurent Balmelli
                     Laboratory For Audio-Visual Communications 
                     Ecole Polytechnique Federale de Lausanne, Switzerland

                     also:
                     Mathematics for Communication, Bell Laboratories
                     Lucent Technologies, USA

  contact, bugs:     Laurent.Balmelli@epfl.ch
                     balmelli@research.bell-labs.com
                     
  PLEASE DO NOT REMOVE THIS COPYRIGHT NOTICE. 
                     
 */


//
// This class group a set of operator to convert image into
// buffers of different types.
//
// Laurent Balmelli, Java Project
//

import java.lang.Math;
import java.awt.*;
import java.awt.image.*;
import java.awt.Toolkit;
import java.lang.reflect.Array;

 class ImageConverter
{

	Image data;
	int   imageWidth,
	      imageHeight;
	int   oddPixelsHeight[],
	      oddPixelsWidth[];

	public static final int ALPHA = 0;
	public static final int RED   = 1;
	public static final int GREEN = 2;
	public static final int BLUE  = 3;	

	public ImageConverter(Image img, int width, int height)
	{
		this.data = img;
		
		imageWidth = width;
		imageHeight = height;	
	}

	public int[] getIntBuffer()
	{
		int[] pixels = new int[imageWidth*imageHeight];

		PixelGrabber pix = new PixelGrabber(data,0,0,
						    imageWidth,imageHeight,
						    pixels,0,imageWidth);

		try {
		   pix.grabPixels();
		} catch (InterruptedException e) {
            		System.err.println("interrupted waiting for pixels!");
            		return null;
        	}
        	if ((pix.status() & ImageObserver.ABORT) != 0) {
            		System.err.println("image fetch aborted or errored");
            		return null;
        	}

		return pixels;
	}

	public double[] getDoubleBufferComponent( int[] pixels, int Component)
	{
		double buffer[] = new double[imageWidth*imageHeight];
		int values[];

		for (int j = 0; j < imageHeight; j++) {
            	  for (int i = 0; i < imageWidth; i++) {
                	values = handleSinglePixel(i, j, pixels[j * imageWidth + i]);
			buffer[j*imageWidth+i] = (double)values[Component];
		  }
		}
		
		return buffer;	
	}
/*
        public void scaleBuffer(int orgx, int orgy, int width, 
				int height, double buffer[], double scale) {
	  
	  for ( int i=orgy; i < height; i++ ) {
	    for ( int j = orgx; j < width; j++ ) {

	      buffer[i*width+j] *= scale;
	    }
          }
        }
*/
	public void scaleBuffer(int orgx, int orgy, int width, 
				int height, double buffer[], double scale) {

	 int shifty = orgy*imageWidth,
	      shiftx,
	      boundx = width-orgx,
	      boundy = height-orgy; 

	 double m = Double.MIN_VALUE;

	  shiftx = imageWidth-(orgx+boundx)+orgx;

	  for ( int i= 0; i < boundy; i++ ) {
	    for ( int j = 0; j < boundx; j++ ) {

		buffer[shifty+j+orgx+i*(imageWidth)] *= scale;
	    }
	  }
	}
/*
	public void translateBuffer(int orgx, int orgy, int width, 
				int height, double buffer[], double trans) {
	  
	  for ( int i=orgy; i < height; i++ ) {
	    for ( int j = orgx; j < width; j++ ) {

	      buffer[i*width+j] += trans;
	    }
          }
        }
*/
	public void translateBuffer(int orgx, int orgy, int width, 
				int height, double buffer[], double trans) {

	 int shifty = orgy*imageWidth,
	      shiftx,
	      boundx = width-orgx,
	      boundy = height-orgy; 

	 double m = Double.MIN_VALUE;

	  shiftx = imageWidth-(orgx+boundx)+orgx;

	  for ( int i= 0; i < boundy; i++ ) {
	    for ( int j = 0; j < boundx; j++ ) {

		buffer[shifty+j+orgx+i*(imageWidth)] += trans;
	    }
	  }
	}	

	// 
	// int to double conversion modes
	//

	public static final int REAL 	= 0;	// do nothing
	public static final int SCALE   = 1;	// scale coefficients
	public static final int ENHANCE = 2;	// enhance colors


	public void enhance(int orgx, int orgy, int width, 
			    int height, 
			    double red[], double green[], double blue[]) {

	  //System.err.println("enhancing square ("+orgx+","+orgy+")-("+width+","+height);


	  //
	  // get minimas
	  //

	   double minr, maxr, ming, maxg, minb, maxb;
	
	   minr = getMin(red, orgx,orgy, width, height);	
	   maxr = getMax(red, orgx,orgy, width, height);
	   ming = getMin(green, orgx,orgy, width, height);	
	   maxg = getMax(green, orgx,orgy, width, height);
           minb = getMin(blue, orgx,orgy, width, height);	
	   maxb = getMax(blue, orgx,orgy, width, height);

	   //System.err.println("minr = "+minr+", ming = "+ming+", minb = "+minb);
	   //System.err.println("maxr = "+maxr+", maxg = "+maxg+", maxb = "+maxb);

	   double mm = Math.min(minr, ming),
	          ml = Math.min(minr, minb),
	          minimum;

           minimum = -Math.min(mm,ml);
		
	   //System.err.println("translate "+minimum);
	 
           translateBuffer( orgx,orgy, width, height,red,minimum);
           translateBuffer( orgx,orgy, width, height,green,minimum);		
           translateBuffer( orgx,orgy, width, height,blue,minimum);
	  	
	   minr = getMin(red, orgx,orgy, width, height);	
	   maxr = getMax(red, orgx,orgy, width, height);
	   ming = getMin(green, orgx,orgy, width, height);	
	   maxg = getMax(green, orgx,orgy, width, height);
           minb = getMin(blue, orgx,orgy, width, height);	
	   maxb = getMax(blue, orgx,orgy, width, height);
	
	   //System.err.println("minr = "+minr+", ming = "+ming+", minb = "+minb);
	   //System.err.println("maxr = "+maxr+", maxg = "+maxg+", maxb = "+maxb);

           double diffr = maxr - minr,
		  diffg = maxg - ming,
		  diffb = maxb - minb,
		  max, f;

           max = Math.max(Math.max(diffr,diffg), Math.max(diffr,diffb));
	   f = 255.0/max;

	   //System.err.println("f="+f);

	   scaleBuffer( orgx,orgy, width, height,red,f);
	   scaleBuffer( orgx,orgy, width, height,green,f);		
           scaleBuffer( orgx,orgy, width, height,blue,f);	

	}	

	public void enhanceBuffer(int width, int height, 
			          int level,
				  double red[], double green[], double blue[]) {

	   if ( level == 0 ) { 	// one single buffer to treat (LL)
		
		enhance(0,0,width,height,red,green,blue);
	   }
	   else  {	// treat the LH, HL, HH

		////System.err.println("enhancing square ("+width/2+","+0+")-("+width+","+height);
		enhance(width/2,0,width,height,red,green,blue);
		////System.err.println("enhancing square ("+0+","+height/2+")-("+width+","+height);
		enhance(0,height/2,width,height,red,green,blue);
		////System.err.println("enhancing square ("+width/2+","+height/2+")-("+width+","+height);
		enhance(width/2,height/2,width,height,red,green,blue);

		enhanceBuffer(width/2, height/2, level-1, red, green, blue);

	   }
	}



	public int[] DoubleBufferToIntBuffer(double alpha[], double red[],
					     double green[], double blue[],
					     int mode, int level)	
	{	
	//
	// int level	this variable is used to target the coefficients enhancing
	//

	   int inverse[] = new int[imageWidth*imageHeight];

	   double copy_red[], copy_green[], copy_blue[];

	   copy_red   = new double[imageWidth*imageHeight];
	   copy_green = new double[imageWidth*imageHeight];
	   copy_blue  = new double[imageWidth*imageHeight];

	   System.arraycopy(red,0,copy_red,0,imageWidth*imageHeight);
	   System.arraycopy(green,0,copy_green,0,imageWidth*imageHeight);
	   System.arraycopy(blue,0,copy_blue,0,imageWidth*imageHeight);
	

	   //System.err.println(" images "+imageWidth+"x"+imageHeight);

	   // Get min and max of all buffers;

	   //	LL LH
	   //   HL HH

	   double minLLr, maxLLr, minLLg, maxLLg, minLLb, maxLLb, 
		  minHLr, maxHLr, minHLg, maxHLg, minHLb, maxHLb,
		  minLHr, maxLHr, minLHg, maxLHg, minLHb, maxLHb,
		  minHHr, maxHHr, minHHg, maxHHg, minHHb, maxHHb;

	   // LL
	   
	   minLLr = getMin(red, 0,0, imageWidth, imageHeight);	
	   maxLLr = getMax(red, 0,0, imageWidth, imageHeight);
	   minLLg = getMin(green, 0,0, imageWidth, imageHeight);	
	   maxLLg = getMax(green, 0,0, imageWidth, imageHeight);
           minLLb = getMin(blue, 0,0, imageWidth, imageHeight);	
	   maxLLb = getMax(blue, 0,0, imageWidth, imageHeight);
          
	   // LH
/*
	   minLHr = getMin(red, imageWidth/2,0,imageWidth,imageHeight/2);
	   maxLHr = getMax(red, imageWidth/2,0,imageWidth,imageHeight/2);
	   minLHg = getMin(green, imageWidth/2,0,imageWidth,imageHeight/2);
	   maxLHg = getMax(green, imageWidth/2,0,imageWidth,imageHeight/2);
	   minLHb = getMin(blue, imageWidth/2,0,imageWidth,imageHeight/2);
	   maxLHb = getMax(blue, imageWidth/2,0,imageWidth,imageHeight/2);

	   // HL
 
	   minHLr = getMin(red, 0,imageHeight/2,imageWidth/2,imageHeight);
	   maxHLr = getMax(red, 0,imageHeight/2,imageWidth/2,imageHeight);
	   minHLg = getMin(green, 0,imageHeight/2,imageWidth/2,imageHeight);
	   maxHLg = getMax(green, 0,imageHeight/2,imageWidth/2,imageHeight);
           minHLb = getMin(blue, 0,imageHeight/2,imageWidth/2,imageHeight);
	   maxHLb = getMax(blue, 0,imageHeight/2,imageWidth/2,imageHeight);
		
	   // HH

	   minHHr = getMin(red, imageWidth/2,imageHeight/2,imageWidth,imageHeight);
	   maxHHr = getMax(red, imageWidth/2,imageHeight/2,imageWidth,imageHeight);
	   minHHg = getMin(green, imageWidth/2,imageHeight/2,imageWidth,imageHeight);
	   maxHHg = getMax(green, imageWidth/2,imageHeight/2,imageWidth,imageHeight);
	   minHHb = getMin(blue, imageWidth/2,imageHeight/2,imageWidth,imageHeight);
	   maxHHb = getMax(blue, imageWidth/2,imageHeight/2,imageWidth,imageHeight);
*/	     	

   //System.err.println("LL: maxr = "+maxLLr+" minr = "+minLLr+" maxg = "+maxLLg+" ming = "+minLLg+"  maxb = "+maxLLb+" minb = "+minLLb);
  // //System.err.println("LH: maxr = "+maxLHr+" minr = "+minLHr+" maxg = "+maxLHg+" ming = "+minLHg+"  maxb = "+maxLHb+" minb = "+minLHb);
  // //System.err.println("HL: maxr = "+maxHLr+" minr = "+minHLr+" maxg = "+maxHLg+" ming = "+minHLg+"  maxb = "+maxHLb+" minb = "+minHLb);
  // //System.err.println("HH: maxr = "+maxHHr+" minr = "+minHHr+" maxg = "+maxHHg+" ming = "+minHHg+"  maxb = "+maxHHb+" minb = "+minHHb);

	   int redc,greenc,bluec;	

	   if ( mode == SCALE ) {
	     scaleBuffer(imageWidth/2,0,imageWidth,imageHeight/2,copy_red,10.0);
	     scaleBuffer(imageWidth/2,imageHeight/2,imageWidth,imageHeight,copy_red,10.0);
	     scaleBuffer(0,imageHeight/2,imageWidth/2,imageHeight,copy_red,10.0);

	     scaleBuffer(imageWidth/2,0,imageWidth,imageHeight/2,copy_green,10.0);
	     scaleBuffer(imageWidth/2,imageHeight/2,imageWidth,imageHeight,copy_green,10.0);
	     scaleBuffer(0,imageHeight/2,imageWidth/2,imageHeight,copy_green,10.0);

	     scaleBuffer(imageWidth/2,0,imageWidth,imageHeight/2,copy_blue,10.0);
	     scaleBuffer(imageWidth/2,imageHeight/2,imageWidth,imageHeight,copy_blue,10.0);
	     scaleBuffer(0,imageHeight/2,imageWidth/2,imageHeight,copy_blue,10.0);
	   }

	   if ( mode == ENHANCE ) {

		/*
		//System.err.println("enhancing colors");

		double mm = Math.min(minLLr, minLLg),
	               ml = Math.min(minLLr, minLLb),
	               minimum;

		minimum = Math.abs(Math.min(mm,ml));
		
		translateBuffer( 0,0, imageWidth, imageHeight,copy_red,minimum);
		translateBuffer( 0,0, imageWidth, imageHeight,copy_green,minimum);		
		translateBuffer( 0,0, imageWidth, imageHeight,copy_blue,minimum);
				
		double diffr = maxLLr - minLLr,
		       diffg = maxLLr - minLLr,
		       diffb = maxLLr - minLLr,
		       max, f;

		max = Math.max(Math.max(diffr,diffg), Math.max(diffr,diffb));
		f = 255.0/max;

		scaleBuffer( 0,0, imageWidth, imageHeight,copy_red,f);
		scaleBuffer( 0,0, imageWidth, imageHeight,copy_green,f);		
		scaleBuffer( 0,0, imageWidth, imageHeight,copy_blue,f);
		*/

		//System.err.println("enhancing buffer level "+level);

		enhanceBuffer(imageWidth,imageHeight,level,copy_red,copy_green,copy_blue);
		
	   }


	   for (int j = 0; j < imageHeight; j++) {
             for (int i = 0; i < imageWidth; i++) {

 		redc = Convert8bits(copy_red[j*imageWidth+i]);
		greenc = Convert8bits(copy_green[j*imageWidth+i]);
		bluec = Convert8bits(copy_blue[j*imageWidth+i]);
	
		 inverse[j*imageWidth+i] = generateSinglePixel(
		 	 (int)(alpha[j*imageWidth+i]),redc,greenc,bluec);
	     }
	   }
	
	   return inverse;     
	 }

/*
	public double getMin(double buffer[], int orgx, int orgy, 
			     int width, int height) {

		double m = Double.MAX_VALUE;

		for ( int i=orgy; i < imageHeight; i++ ) {
	    	   for ( int j = orgx; j < imageWidth; j++ ) {

			if ( buffer[i*imageWidth+j] < m )
			    m = buffer[i*imageWidth+j];
	    	   }
          	}
		
		return m;
	}
*/
/*
       public double getMin(double buffer[], int orgx, int orgy, 
			     int width, int height) {
	 int shifty = orgy*imageWidth,
	      shiftx,
	      boundx = width-orgx,
	      boundy = height-orgy; 

	 double m = Double.MAX_VALUE;

	  shiftx = imageWidth-(orgx+boundx)+orgx;

	  for ( int i= 0; i < boundy; i++ ) {
	    for ( int j = 0; j < boundx; j++ ) {

		if ( buffer[shifty+orgx+i*(boundx+shiftx)+j] < m )
		  m = buffer[shifty+orgx+i*(boundx+shiftx)+j];
	    }
	  }

	  return m;
	}
*/

	public double getMin(double buffer[], int orgx, int orgy, 
			     int width, int height) {
	 int shifty = orgy*imageWidth,
	      shiftx,
	      boundx = width-orgx,
	      boundy = height-orgy; 

	  double m = Double.MAX_VALUE;

	  shiftx = imageWidth-(orgx+boundx);

	  for ( int i= 0; i < boundy; i++ ) {
	    for ( int j = 0; j < boundx; j++ ) {
				
		if ( buffer[shifty+j+orgx+i*(imageWidth)] < m ) 
		   m = buffer[shifty+j+orgx+i*(imageWidth)];
		
	    }
	  }

	  return m;
	}


	public double getMax(double buffer[], int orgx, int orgy, 
			     int width, int height) {
	 int shifty = orgy*imageWidth,
	      shiftx,
	      boundx = width-orgx,
	      boundy = height-orgy; 

	 double m = Double.MIN_VALUE;

	  shiftx = imageWidth-(orgx+boundx)+orgx;

	  for ( int i= 0; i < boundy; i++ ) {
	    for ( int j = 0; j < boundx; j++ ) {

		if ( buffer[shifty+j+orgx+i*(imageWidth)] > m )
		  m = buffer[shifty+j+orgx+i*(imageWidth)];
	    }
	  }
	 
	  return m;
	}

/*
	public double getMax(double buffer[], int orgx, int orgy, 
			     int width, int height) {

		double m = Double.MIN_VALUE;

		for ( int i=orgy; i < imageHeight; i++ ) {
	    	   for ( int j = orgx; j < imageWidth; j++ ) {

			if ( buffer[i*imageWidth+j] > m )
			    m = buffer[i*imageWidth+j];
	    	   }
          	}

		return m;
	}
*/
	

	
/*
	public int Convert8bits( double value )
	{
	  
	   if ( ( value > 255) || ( value < 0 ) ) {
	        if ( value > 255 ) return 255;
	 	if ( value < 0 ) return 0;
	   }
	    else
	  	return (int)value;

	   return 0;
	}
*/
	public int Convert8bits( double value )
	{
	  
	   if ( ( value > 255) || ( value < 0 ) ) {
	        if ( value > 255 ) return 255;
	 	if ( value < 0 ) return 0;
	   }
	    else
	  	return (int)value;

	   return 0;
	}
	

	public double[] transposeBuffer(double buffer[],int width,int height)
	{
	    double result[] = new double[width*height];

	    //System.out.println("size is "+imageWidth+"x"+imageHeight);
	    for (int j = 0; j < height; j++) {
             for (int i = 0; i < width; i++) {
		//System.out.println("result("+i+","+j*imageWidth+") = buffer ("+j+","+i*imageWidth+")");
		result[j + i*height] = buffer[j*width+i];
	     }
	    }

	    return result;
	}


	// TO DO
	// Terminate this routine!!!
/*
	public void convertEvenLength()
	{
	   int newHeight,
	       newWidth;
	   
	   boolean changeHeight = false,
		   changeWidth  = false;
		
	   if ( ( height%2 ) != 0 )
	   {	
		changeHeight = true;
		newHeight = height-1;
	        System.out.println("Height has been updated.");
	   }
	   
	   if ( ( width%2 ) != 0 )
	   {
		changeWidth = true;
		newWidth = width-1;
	        System.out.println("Height has been updated.");
	   }   
	}	
*/
	// --------------------------------------------------------------------	
	// Private methods
	// --------------------------------------------------------------------

	private int[] handleSinglePixel(int x, int y, int pixel) {

        	int alpha = (pixel >> 24) & 0xff;
        	int red   = (pixel >> 16) & 0xff;
        	int green = (pixel >>  8) & 0xff;
        	int blue  = (pixel      ) & 0xff;

		int values[] = new int[4];

		values[0] = alpha; values[1] = red; values[2] = green; values[3] = blue;

		return values;
        }

	private int generateSinglePixel(int values[]) {
	// 
	// <values> must have format 0..3: ALPHA:RED:GREEN:BLUE
	//
		int result;
		result = 0;	// clear the vector

	 	result  = values[0]; result = (result << 8);  // store ALPHA channel
		result += values[1]; result = (result << 8);  // store RED channel
		result += values[2]; result = (result << 8);  // store GREEN channel
		result += values[3];  			      // store BLUE channel
		
		return result;

	}

	private int generateSinglePixel(int alpha, int red, int green, int blue) {
	// 
	// 
	//
		int result;
		result = 0;	// clear the vector

	 	result  = alpha; result = (result << 8);  // store ALPHA channel
		result += red;   result = (result << 8);  // store RED channel
		result += green; result = (result << 8);  // store GREEN channel
		result += blue;  			  // store BLUE channel
		
		return result;

	}
}
