/*

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish and distribute.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
  AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Copyright 1999 -   Laurent Balmelli
                     Laboratory For Audio-Visual Communications 
                     Ecole Polytechnique Federale de Lausanne, Switzerland

                     also:
                     Mathematics for Communication, Bell Laboratories
                     Lucent Technologies, USA

  contact, bugs:     Laurent.Balmelli@epfl.ch
                     balmelli@research.bell-labs.com
                     
  PLEASE DO NOT REMOVE THIS COPYRIGHT NOTICE. 
                     
 */

import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.*;
//import java.lang.*;
import java.lang.Math;
import java.lang.Double;
import java.util.Vector;
import java.lang.Object;

//import java.io.*;


/**
 * This class implements a set of tools to plot	a signalS
 * on the screen.
 *
 */

//
//	Resize of the image
//
//

class ImageDisplay extends Panel //implements java.io.Serializable 
{


	//	
	// Object State		
	//

	private Color 	BackgroundColor	= Color.white;
	private Color 	boxColor 	= Color.black;
	private Color 	gridColor 	= Color.lightGray;
	private int	MAXIMAGES	= 10;

	private Vector  Anotations	= new Vector();
	

	private boolean	haveimage;

	private double 	xScale, 
			yScale;						// signal value to pixels scalings
	private int 	xOrg, 
			yOrg;						// display's (0,0) coordinates

	private int 	canvasWidth, 
			canvasHeight;					// outer panel's dimensions
	
	//private Insets 	boxMargins = new Insets(30, 40, 30, 40);	// Distance of the drawing area 
									// from the borders of the outer panel:		
		private Insets 	boxMargins = new Insets(30, 4, 3, 4);

	public boolean autoScale = true;
	/**
	 * Controls labeling of the axis.
	 *
	 * If set to true, numeric labels are added to the axis of the plot.
	 */
	public int autoLabels = XYON;
	
	private Font 	labelsFont = new Font("Helvetica", Font.PLAIN, 12); // Standard settings for the graphics:
	private Font 	titleFont = new Font("Helvetica", Font.PLAIN, 14);
	
	private boolean	DrawZoom;

 	private Vector 	images;
	private int	index;
	public  int	displayIndex;

	private int 	imageWidth[];
	private int	imageHeight[];

	public String titleString = "";

	private int height = 200;
	private int width = 250;
	
	private final int ZOOM_LEVEL = 50;
	private int ZoomLevel = 0;
	private int MaxZoom = 0;

	private double xMin[],
		       xMax[],
		       yMin[],
		       yMax[];

	public final static int XYON = 0;
	public final static int ON = 0;
	public final static int XYOFF = 1;
	public final static int OFF = 1;
	public final static int XOFF = 2;
	public final static int YOFF = 3;  

	public boolean BOX;

	MediaTracker tracker;
	URL url_image;


	/**
	 * Create a Canvas which can accomodate a plotting area of <tt> w X h </tt>
	 * pixels. The background color is set to black and the foreground to white
	 * unchangeable by now).
	 *
	 * @param w		the width of the drawing area
	 * @param h 	the height of the drawing area
	 *
	 * @see			dsp.Signal
	 * @see			dsp.BorderPanel
	 */ 	 
	public ImageDisplay(int w, int h) {

		//System.err.println("Creating Signal Display");

		BackgroundColor = Color.white;
		
		BOX = true;	// draw a nice box by default;

		height = h;
		width = w;

		canvasWidth = width + boxMargins.left + boxMargins.right;
		canvasHeight = height + boxMargins.top + boxMargins.bottom; 

		xMin = new double[ZOOM_LEVEL];
		xMax = new double[ZOOM_LEVEL];
		yMin = new double[ZOOM_LEVEL];
		yMax = new double[ZOOM_LEVEL];

		for ( int i = 0; i < ZOOM_LEVEL; i++ ) {
		   xMin[i] = 0.0; xMax[i] = 100.0;
		   yMin[i] = 0.0; yMax[i] = 100.0;
	        }	
		
		imageWidth = new int[MAXIMAGES];
		imageHeight = new int[MAXIMAGES];
		images = new Vector(MAXIMAGES);
	
		index = -1;
		haveimage = false;
		
	}

	public Dimension getMinimumSize() {
	   return new Dimension(canvasWidth,canvasHeight);
		}
	
	public Dimension getPreferredSize() {
	   return new Dimension(canvasWidth,canvasHeight);
	}

	public Dimension maximumSize() {
	   return new Dimension(canvasWidth,canvasHeight);
	}
	

	/**
	 * Set the background color of the canvas
	 *
	 * @param c		the color for the backgournd
	 *
	 * @see			awt.Color
	 */

 	public void SetBackgroundColor(Color c) {
		
		BackgroundColor = c;

		if ( c == Color.black )	{
			boxColor = Color.white;
		} 
	}



	/**
	 * Plot a signal on the current display.
	 *
	 * @param signal	the signal to be plotted.
	 *
	 * @see				dsp.SignalDisplay#autoScale
	 */
	/*
	 * The set of signals to be plotted is implemented as a linked list.
	 * See also the private class SignalList.
	 */

	public void plot(Image i) {

		index++;
		displayIndex = index;

		images.addElement(i);

		System.err.println(".");

		imageWidth[index]  = i.getHeight(this); 
		imageHeight[index] = i.getWidth(this);  
		
		System.err.println("w ="+imageWidth[index]);		

		ZoomLevel = 0;
		haveimage = true;
		repaint();
	}

	public void erase(int i) {
		
	   if ( (i < 0) || ( i > index) || ( images.size() == 0 ))
	      error(OUT_OF_BOUNDS);
	   else
	      images.removeElementAt(i);

	   if ( index > -1 )	// -1 is the initial value
	      index--;
		
	   displayIndex = index;

	   if ( images.size() == 0 ) haveimage = false;
	 
        }	

	public void ZoomInFigure() {
		if ( ZoomLevel < MaxZoom )
		   ZoomLevel += 1;

		repaint();
	}

	public void ZoomOutFigure() {
		if ( ZoomLevel > 0 )
		   ZoomLevel -= 1;

		 repaint();
	}	
		   
	
	/**
	 * Clears the current display.
	 *
	 * @see				dsp.SignalDisplay#autoScale
	 * @see				dsp.SignalDisplay#autoLabels
	 */
	public void cls() {
	
		repaint();
	}	
		
	/* 
	 * Following are a set of private helpers to draw various things on
	 * the current display.
	 */
	 

	
	// variables used in drawing dashed and dotted lines:
	private double dashRemainder;
	private boolean dashDraw;


	/**
	 * Zoom on a signal on the current display.
	 *
	 * @param  
	 *
	 * 
	 */
	 	
	public void zoom(int topX, int topY, int bottomX, int bottomY) {


		// compute coefficients for the affine transformation
	
		//System.err.println(topX+" "+topY+" "+bottomX+" "+bottomY);

		double alphaX = Math.abs((bottomX-topX))/(double)width;
		double alphaY = Math.abs((bottomY-topY))/(double)height;
  
		int betaX = (topX-boxMargins.left);
		int betaY = (topY-boxMargins.top);	
	
		// convert coefficient in function space
		
		double LengthX = xMax[ZoomLevel] - xMin[ZoomLevel],
		       LengthY = yMax[ZoomLevel] - yMin[ZoomLevel];

		double unitX = LengthX/width,
		       unitY = LengthY/height;

 
		double limitXmin = xMin[ZoomLevel]+(topX-boxMargins.left)*unitX,	 
		       limitXmax = xMin[ZoomLevel]+(bottomX-boxMargins.left)*unitX,  
		       limitYmax = yMin[ZoomLevel]-(topY-height-boxMargins.top)*unitY,  
		       limitYmin = yMin[ZoomLevel]-(bottomY-height-boxMargins.top)*unitY;   

	
		ZoomLevel+=1;
		MaxZoom = ZoomLevel;
		 
		xMin[ZoomLevel] = limitXmin; xMax[ZoomLevel] = limitXmax; 
		yMin[ZoomLevel] = limitYmin; yMax[ZoomLevel] = limitYmax;
		
		DrawZoom = false;

		repaint();	 	
	}


	/*
	 * This draws a white box around the drawing area, together with 
	 * title, axis labels and grid (if required).
	 */
	private void drawBox(Graphics g) {

		setBackground(BackgroundColor);
		/*
		if ( (imageHeight[index] == 0) && (imageWidth[index] == 0) ) {
		   xMin[ZoomLevel] = 0; xMax[ZoomLevel] = 100;  
		   yMin[ZoomLevel] = 0; yMax[ZoomLevel] = 100;
		}
		*/
		xOrg = boxMargins.left;
		yOrg = height + boxMargins.top;

		g.setFont(titleFont);

		FontMetrics fm = g.getFontMetrics();

		g.drawString(titleString, 
			xOrg + (width - fm.stringWidth(titleString))/2, 
			yOrg - height - fm.getHeight()/2);
		
		// draw the box and clip the area so that lines don't spill ove
		
		if ( BOX )	
		   g.drawRect(xOrg-1, yOrg -height-1, width+1, height+2);

		g.clipRect(xOrg, yOrg-height, width, height);
	}
	
	
	public void paint(Graphics g) {

		drawBox(g);
		
		if ( haveimage ) {
			
			if ( (displayIndex > index) || (displayIndex < 0) ) 
		   		error(OUT_OF_BOUNDS);
			else		
		   		g.drawImage((Image)(images.elementAt(displayIndex)),boxMargins.left,boxMargins.top,this);    	
		}

		else {
			g.setColor(BackgroundColor);
			g.fillRect(xOrg, yOrg-height, width, height+1);

		}
	}


/*	public Dimension getMinimumSize() {
		return new Dimension(canvasWidth, canvasHeight);
	}
	public Dimension getPreferredSize() {
		return getMinimumSize();
	}*/

	/******************************************************************************
		Errors Handling
	******************************************************************************/

	private static final int SIGNAL_SIZE   = 0;
	private static final int OUT_OF_BOUNDS = 1;

	private static final 
		String SizeError = new String("??? Error using ==> plot\n Vectors must be the same lengths.");

	private static final 
		String OutOfBoundError = new String("Error: image index out of bounds, or no image to erase.");

	private void error(int type) {

	   switch(type) {
	
		case SIGNAL_SIZE:
		  System.err.println(SizeError);
		break;

		case OUT_OF_BOUNDS:
		  System.err.println(OutOfBoundError);
		break;
	   }			  
	
	}

	/******************************************************************************
		Image Anotation
	******************************************************************************/
/*
	public void insertAnotation(Anotation) {


	}	
*/
	
}
	

